"""Provide the exceptions for the plugin manager."""

from fastapi_factory_utilities.core.exceptions import FastAPIFactoryUtilitiesError


class PluginManagerError(FastAPIFactoryUtilitiesError):
    """Generic plugin manager error."""


class InvalidPluginError(PluginManagerError):
    """The plugin is invalid."""

    def __init__(self, plugin_name: str, message: str) -> None:
        """Instantiate the exception.

        Args:
            plugin_name (str): The plugin name.
            message (str): The message
        """
        super().__init__(message=f"Invalid plugin: {plugin_name}, {message}")


class PluginPreConditionNotMetError(PluginManagerError):
    """The plugin pre-condition is not met."""

    def __init__(self, plugin_name: str, message: str) -> None:
        """Instantiate the exception.

        Args:
            plugin_name (str): The plugin name.
            message (str): The message
        """
        super().__init__(message=f"Plugin pre-condition not met: {plugin_name}, {message}")



"""Plugin manager.

This module provides the plugin manager for the application.

Objective:
    - Provide a way to manage plugins for the application.
    - The plugins are activated based on the configuration.
    - The plugins are checked for pre-conditions.
    - The plugins are loaded on application load.
    - The plugins are started on application startup.
    - The plugins are stopped on application shutdown.
"""

from importlib import import_module
from types import ModuleType
from typing import Self

from fastapi_factory_utilities.core.plugins import PluginsEnum, PluginState
from fastapi_factory_utilities.core.protocols import (
    ApplicationAbstractProtocol,
    PluginProtocol,
)

from .exceptions import InvalidPluginError, PluginPreConditionNotMetError


class PluginManager:
    """Plugin manager for the application."""

    DEFAULT_PLUGIN_PACKAGE: str = "fastapi_factory_utilities.core.plugins"

    @classmethod
    def _import_module(cls, name: str, package: str) -> ModuleType:
        """Import the module. This is a wrapper around the import_module function to be able to mock it in the tests.

        Args:
            name (str): The name of the module.
            package (str): The package of the module.

        Returns:
            ModuleType: The module.

        Raises:
            ImportError: If the module cannot be imported.
            ModuleNotFoundError: If the module is not found.
        """
        return import_module(name=f"{package}.{name}")

    @classmethod
    def _check_pre_conditions(
        cls, plugin_package: str, want_to_activate_plugins: list[PluginsEnum], application: ApplicationAbstractProtocol
    ) -> list[PluginProtocol]:
        """Check the pre-conditions for the plugins.

        Args:
            plugin_package (str): The package for the plugins.
            want_to_activate_plugins (list[PluginsEnum]): The plugins to activate.
            application (BaseApplicationProtocol): The application.

        Returns:
            list[PluginProtocol]: The activated plugins.

        Raises:
            InvalidPluginError: If the plugin is invalid.
            PluginPreConditionNotMetError: If the pre-conditions are not met.
        """
        plugins: list[PluginProtocol] = []

        for plugin_enum in want_to_activate_plugins:

            try:
                # Using a custom import function to be able to mock it in the tests.
                plugin_module: ModuleType = cls._import_module(name=plugin_enum.value, package=plugin_package)
            except (ImportError, ModuleNotFoundError) as import_error:
                # TODO: Later, if we introduce extra mecanism to manage dependencies,
                # we must handle a part of the error here. To be able to provide a better error message.
                # For now, we just raise the error.
                raise InvalidPluginError(
                    plugin_name=plugin_enum.value, message="Error importing the plugin "
                ) from import_error

            if not isinstance(plugin_module, PluginProtocol):
                raise InvalidPluginError(
                    plugin_name=plugin_enum.value, message="The plugin does not implement the PluginProtocol"
                )

            if not plugin_module.pre_conditions_check(application=application):
                raise PluginPreConditionNotMetError(
                    plugin_name=plugin_enum.value, message="The pre-conditions are not met"
                )

            plugins.append(plugin_module)

        return plugins

    def __init__(self, activated: list[PluginsEnum] | None = None, plugin_package: str | None = None) -> None:
        """Instanciate the plugin manager."""
        self._plugin_package: str = plugin_package or self.DEFAULT_PLUGIN_PACKAGE
        self._plugins_wanted_to_be_activated: list[PluginsEnum] = activated or []
        self._activated_plugins: list[PluginProtocol] = []
        self._states: list[PluginState] = []

    def activate(self, plugin: PluginsEnum) -> Self:
        """Activate a plugin.

        Args:
            plugin (PluginsEnum): The plugin to activate.

        Returns:
            Self: The plugin manager.
        """
        self._plugins_wanted_to_be_activated.append(plugin)

        return self

    @property
    def states(self) -> list[PluginState]:
        """Get the states."""
        return self._states

    def add_application_context(self, application: ApplicationAbstractProtocol) -> Self:
        """Add the application context to the plugins.

        Args:
            application (BaseApplicationProtocol): The application.

        Returns:
            Self: The plugin manager.
        """
        self._application: ApplicationAbstractProtocol = application

        return self

    def add_states(self, states: list[PluginState]) -> Self:
        """Add the states to the plugin manager.

        Args:
            states (list[PluginState]): The states.

        Returns:
            Self: The plugin manager.
        """
        for state in states:
            self._states.append(state)
        return self

    def clear_states(self) -> Self:
        """Clear the states."""
        self._states = []
        return self

    def load(self) -> Self:
        """Load the plugins.

        Returns:
            Self: The plugin manager.

        Raises:
            InvalidPluginError: If the plugin is invalid.
            PluginPreConditionNotMetError: If the pre-conditions are not met.

        """
        # Remove duplicates.
        self._plugins_wanted_to_be_activated = list(set(self._plugins_wanted_to_be_activated))
        # Check the pre-conditions for the plugins.
        self._activated_plugins = self._check_pre_conditions(
            plugin_package=self._plugin_package,
            want_to_activate_plugins=self._plugins_wanted_to_be_activated,
            application=self._application,
        )
        # Load from the previous check.
        for plugin in self._activated_plugins:
            self.add_states(states=plugin.on_load(application=self._application) or [])

        return self

    async def trigger_startup(self) -> Self:
        """Trigger the startup of the plugins."""
        for plugin in self._activated_plugins:
            states: list[PluginState] | None = await plugin.on_startup(application=self._application)
            self.add_states(states=states or [])

        return self

    async def trigger_shutdown(self) -> Self:
        """Trigger the shutdown of the plugins."""
        for plugin in self._activated_plugins:
            await plugin.on_shutdown(application=self._application)

        return self



"""Httpx Plugin Module."""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from fastapi_factory_utilities.core.plugins import PluginState
    from fastapi_factory_utilities.core.protocols import ApplicationAbstractProtocol


def pre_conditions_check(application: "ApplicationAbstractProtocol") -> bool:  # pylint: disable=unused-argument
    """Check the pre-conditions for the example plugin."""
    return True


def on_load(
    application: "ApplicationAbstractProtocol",  # pylint: disable=unused-argument
) -> list["PluginState"] | None:
    """Actions to perform on load for the example plugin."""
    return


async def on_startup(
    application: "ApplicationAbstractProtocol",  # pylint: disable=unused-argument
) -> list["PluginState"] | None:
    """Actions to perform on startup for the example plugin."""
    return


async def on_shutdown(application: "ApplicationAbstractProtocol") -> None:  # pylint: disable=unused-argument
    """Actions to perform on shutdown for the example plugin."""
    return



"""Provides the core application module for the Python Factory."""

from .application import ApplicationAbstract
from .builder import ApplicationGenericBuilder
from .config import BaseApplicationConfig, RootConfig
from .enums import EnvironmentEnum

__all__: list[str] = [
    "BaseApplicationConfig",
    "EnvironmentEnum",
    "ApplicationAbstract",
    "ApplicationGenericBuilder",
    "RootConfig",
]



"""FastAPI Factory Utilities exceptions."""

import logging
from typing import Any

from opentelemetry.trace import Span, get_current_span
from structlog.stdlib import BoundLogger, get_logger

_logger: BoundLogger = get_logger()


class FastAPIFactoryUtilitiesError(Exception):
    """Base exception for the FastAPI Factory Utilities."""

    def __init__(
        self,
        *args: tuple[Any],
        message: str | None = None,
        level: int = logging.ERROR,
        **kwargs: dict[str, Any],
    ) -> None:
        """Instanciate the exception.

        Args:
            *args (Tuple[Any]): The arguments.
            message (str | None): The message.
            level (int): The logging level.
            **kwargs (dict[str, Any]): The keyword arguments
        """
        # Log the Exception
        if message:
            _logger.log(level=level, event=message)
            self.message = message
            self.level = level
            args = (message, *args)  # type: ignore
        # Propagate the exception
        span: Span = get_current_span()
        # If not otel is setup, INVALID_SPAN is retrive from get_current_span
        # and it will respond False to the is_recording method
        if span.is_recording():
            span.record_exception(self)
        # Call the parent class
        super().__init__(*args, **kwargs)



"""Protocols for the base application."""

from abc import abstractmethod
from typing import TYPE_CHECKING, ClassVar, Protocol, runtime_checkable

from beanie import Document
from fastapi import FastAPI

from fastapi_factory_utilities.core.plugins import PluginsEnum
from fastapi_factory_utilities.core.services.status.services import StatusService

if TYPE_CHECKING:
    from fastapi_factory_utilities.core.app.config import RootConfig
    from fastapi_factory_utilities.core.plugins import PluginState


class ApplicationAbstractProtocol(Protocol):
    """Protocol for the base application."""

    PACKAGE_NAME: ClassVar[str]

    ODM_DOCUMENT_MODELS: ClassVar[list[type[Document]]]

    DEFAULT_PLUGINS_ACTIVATED: ClassVar[list[PluginsEnum]]

    @abstractmethod
    def get_config(self) -> "RootConfig":
        """Get the application configuration."""

    @abstractmethod
    def get_asgi_app(self) -> FastAPI:
        """Get the ASGI application."""

    @abstractmethod
    def get_status_service(self) -> StatusService:
        """Get the status service."""


@runtime_checkable
class PluginProtocol(Protocol):
    """Defines the protocol for the plugins."""

    @abstractmethod
    def pre_conditions_check(self, application: ApplicationAbstractProtocol) -> bool:
        """Check the pre-conditions for the plugin.

        Args:
            application (BaseApplicationProtocol): The application.

        Returns:
            bool: True if the pre-conditions are met, False otherwise.
        """

    @abstractmethod
    def on_load(self, application: ApplicationAbstractProtocol) -> list["PluginState"] | None:
        """The actions to perform on load for the plugin.

        Args:
            application (BaseApplicationProtocol): The application.

        Returns:
            None
        """

    @abstractmethod
    async def on_startup(self, application: ApplicationAbstractProtocol) -> list["PluginState"] | None:
        """The actions to perform on startup for the plugin.

        Args:
            application (BaseApplicationProtocol): The application.

        Returns:
            None
        """

    @abstractmethod
    async def on_shutdown(self, application: ApplicationAbstractProtocol) -> None:
        """The actions to perform on shutdown for the plugin.

        Args:
            application (BaseApplicationProtocol): The application.

        Returns:
            None
        """



"""Oriented Data Model (ODM) plugin package."""

from logging import INFO, Logger, getLogger
from typing import Any

from beanie import init_beanie  # pyright: ignore[reportUnknownVariableType]
from motor.motor_asyncio import AsyncIOMotorClient
from reactivex import Subject
from structlog.stdlib import BoundLogger, get_logger

from fastapi_factory_utilities.core.plugins import PluginState
from fastapi_factory_utilities.core.protocols import ApplicationAbstractProtocol
from fastapi_factory_utilities.core.services.status.enums import (
    ComponentTypeEnum,
    HealthStatusEnum,
    ReadinessStatusEnum,
)
from fastapi_factory_utilities.core.services.status.services import StatusService
from fastapi_factory_utilities.core.services.status.types import (
    ComponentInstanceType,
    Status,
)

from .builder import ODMBuilder
from .depends import depends_odm_client, depends_odm_database
from .documents import BaseDocument
from .exceptions import OperationError, UnableToCreateEntityDueToDuplicateKeyError
from .repositories import AbstractRepository

_logger: BoundLogger = get_logger()


def pre_conditions_check(application: ApplicationAbstractProtocol) -> bool:
    """Check the pre-conditions for the OpenTelemetry plugin.

    Args:
        application (BaseApplicationProtocol): The application.

    Returns:
        bool: True if the pre-conditions are met, False otherwise.
    """
    del application
    return True


def on_load(
    application: ApplicationAbstractProtocol,
) -> list["PluginState"] | None:
    """Actions to perform on load for the OpenTelemetry plugin.

    Args:
        application (BaseApplicationProtocol): The application.
    """
    del application
    # Configure the pymongo logger to INFO level
    pymongo_logger: Logger = getLogger("pymongo")
    pymongo_logger.setLevel(INFO)
    _logger.debug("ODM plugin loaded.")


async def on_startup(
    application: ApplicationAbstractProtocol,
) -> list["PluginState"] | None:
    """Actions to perform on startup for the ODM plugin.

    Args:
        application (BaseApplicationProtocol): The application.
        odm_config (ODMConfig): The ODM configuration.

    Returns:
        None
    """
    states: list[PluginState] = []

    status_service: StatusService = application.get_status_service()
    component_instance: ComponentInstanceType = ComponentInstanceType(
        component_type=ComponentTypeEnum.DATABASE, identifier="MongoDB"
    )
    monitoring_subject: Subject[Status] = status_service.register_component_instance(
        component_instance=component_instance
    )

    try:
        odm_factory: ODMBuilder = ODMBuilder(application=application).build_all()
        await odm_factory.wait_ping()
    except Exception as exception:  # pylint: disable=broad-except
        _logger.error(f"ODM plugin failed to start. {exception}")
        # TODO: Report the error to the status_service
        # this will report the application as unhealthy
        monitoring_subject.on_next(
            value=Status(health=HealthStatusEnum.UNHEALTHY, readiness=ReadinessStatusEnum.NOT_READY)
        )
        return states

    if odm_factory.odm_database is None or odm_factory.odm_client is None:
        _logger.error(
            f"ODM plugin failed to start. Database: {odm_factory.odm_database} - " f"Client: {odm_factory.odm_client}"
        )
        # TODO: Report the error to the status_service
        # this will report the application as unhealthy
        monitoring_subject.on_next(
            value=Status(health=HealthStatusEnum.UNHEALTHY, readiness=ReadinessStatusEnum.NOT_READY)
        )
        return states

    # Add the ODM client and database to the application state
    states.append(
        PluginState(key="odm_client", value=odm_factory.odm_client),
    )
    states.append(
        PluginState(
            key="odm_database",
            value=odm_factory.odm_database,
        ),
    )

    # TODO: Find a better way to initialize beanie with the document models of the concrete application
    # through an hook in the application, a dynamis import ?
    try:
        await init_beanie(
            database=odm_factory.odm_database,
            document_models=application.ODM_DOCUMENT_MODELS,
        )
    except Exception as exception:  # pylint: disable=broad-except
        _logger.error(f"ODM plugin failed to start. {exception}")
        # TODO: Report the error to the status_service
        # this will report the application as unhealthy
        monitoring_subject.on_next(
            value=Status(health=HealthStatusEnum.UNHEALTHY, readiness=ReadinessStatusEnum.NOT_READY)
        )
        return states

    _logger.info(
        f"ODM plugin started. Database: {odm_factory.odm_database.name} - "
        f"Client: {odm_factory.odm_client.address} - "
        f"Document models: {application.ODM_DOCUMENT_MODELS}"
    )

    monitoring_subject.on_next(value=Status(health=HealthStatusEnum.HEALTHY, readiness=ReadinessStatusEnum.READY))

    return states


async def on_shutdown(application: ApplicationAbstractProtocol) -> None:
    """Actions to perform on shutdown for the ODM plugin.

    Args:
        application (BaseApplicationProtocol): The application.

    Returns:
        None
    """
    # Skip if the ODM plugin was not started correctly
    if not hasattr(application.get_asgi_app().state, "odm_client"):
        return

    client: AsyncIOMotorClient[Any] = application.get_asgi_app().state.odm_client
    client.close()
    _logger.debug("ODM plugin shutdown.")


__all__: list[str] = [
    "BaseDocument",
    "AbstractRepository",
    "OperationError",
    "UnableToCreateEntityDueToDuplicateKeyError",
    "depends_odm_client",
    "depends_odm_database",
]



"""OpenTelemetry Plugin Module."""

import asyncio
from typing import cast

from opentelemetry.instrumentation.fastapi import (  # pyright: ignore[reportMissingTypeStubs]
    FastAPIInstrumentor,
)
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.trace import TracerProvider
from structlog.stdlib import BoundLogger, get_logger

from fastapi_factory_utilities.core.protocols import ApplicationAbstractProtocol

from .builder import OpenTelemetryPluginBuilder
from .configs import OpenTelemetryConfig
from .exceptions import OpenTelemetryPluginBaseException, OpenTelemetryPluginConfigError

__all__: list[str] = [
    "OpenTelemetryConfig",
    "OpenTelemetryPluginBaseException",
    "OpenTelemetryPluginConfigError",
    "OpenTelemetryPluginBuilder",
]

_logger: BoundLogger = get_logger()


def pre_conditions_check(application: ApplicationAbstractProtocol) -> bool:
    """Check the pre-conditions for the OpenTelemetry plugin.

    Args:
        application (BaseApplicationProtocol): The application.

    Returns:
        bool: True if the pre-conditions are met, False otherwise.
    """
    del application
    return True


def on_load(
    application: ApplicationAbstractProtocol,
) -> None:
    """Actions to perform on load for the OpenTelemetry plugin.

    Args:
        application (BaseApplicationProtocol): The application.
    """
    # Build the OpenTelemetry Resources, TracerProvider and MeterProvider
    try:
        otel_builder: OpenTelemetryPluginBuilder = OpenTelemetryPluginBuilder(application=application).build_all()
    except OpenTelemetryPluginBaseException as exception:
        _logger.error(f"OpenTelemetry plugin failed to start. {exception}")
        return
    # Configuration is never None at this point (checked in the builder and raises an exception)
    otel_config: OpenTelemetryConfig = cast(OpenTelemetryConfig, otel_builder.config)
    # Save as state in the FastAPI application
    application.get_asgi_app().state.tracer_provider = otel_builder.tracer_provider
    application.get_asgi_app().state.meter_provider = otel_builder.meter_provider
    application.get_asgi_app().state.otel_config = otel_config
    # Instrument the FastAPI application
    FastAPIInstrumentor.instrument_app(  # pyright: ignore[reportUnknownMemberType]
        app=application.get_asgi_app(),
        tracer_provider=otel_builder.tracer_provider,
        meter_provider=otel_builder.meter_provider,
        excluded_urls=otel_config.excluded_urls,
    )

    _logger.debug(f"OpenTelemetry plugin loaded. {otel_config.activate=}")


async def on_startup(
    application: ApplicationAbstractProtocol,
) -> None:
    """Actions to perform on startup for the OpenTelemetry plugin.

    Args:
        application (BaseApplicationProtocol): The application.

    Returns:
        None
    """
    del application
    _logger.debug("OpenTelemetry plugin started.")


async def on_shutdown(application: ApplicationAbstractProtocol) -> None:
    """Actions to perform on shutdown for the OpenTelemetry plugin.

    Args:
        application (BaseApplicationProtocol): The application.

    Returns:
        None
    """
    tracer_provider: TracerProvider = application.get_asgi_app().state.tracer_provider
    meter_provider: MeterProvider = application.get_asgi_app().state.meter_provider
    otel_config: OpenTelemetryConfig = application.get_asgi_app().state.otel_config

    seconds_to_ms_multiplier: int = 1000

    async def close_tracer_provider() -> None:
        """Close the tracer provider."""
        tracer_provider.force_flush(timeout_millis=otel_config.closing_timeout * seconds_to_ms_multiplier)
        # No Delay for the shutdown of the tracer provider
        tracer_provider.shutdown()

    async def close_meter_provider() -> None:
        """Close the meter provider.

        Split the timeout in half for the flush and shutdown.
        """
        meter_provider.force_flush(timeout_millis=int(otel_config.closing_timeout / 2) * seconds_to_ms_multiplier)
        meter_provider.shutdown(timeout_millis=int(otel_config.closing_timeout / 2) * seconds_to_ms_multiplier)

    _logger.debug("OpenTelemetry plugin stop requested. Flushing and closing...")

    await asyncio.gather(
        close_tracer_provider(),
        close_meter_provider(),
    )

    _logger.debug("OpenTelemetry plugin closed.")



"""Provide Kratos Session and Identity classes."""

from typing import Annotated

from fastapi import Depends, HTTPException, Request

from fastapi_factory_utilities.core.services.kratos import (
    KratosOperationError,
    KratosService,
    KratosSessionInvalidError,
    KratosSessionObject,
    depends_kratos_service,
)


class KratosSessionAuthentication:
    """Kratos Session class."""

    DEFAULT_COOKIE_NAME: str = "ory_kratos_session"

    def __init__(self, cookie_name: str = DEFAULT_COOKIE_NAME) -> None:
        """Initialize the KratosSessionAuthentication class.

        Args:
            cookie_name (str): Name of the cookie to extract the session
        """
        self._cookie_name: str = cookie_name

    def _extract_cookie(self, request: Request) -> str:
        """Extract the cookie from the request.

        Args:
            request (Request): FastAPI request object.

        Returns:
            str | None: Cookie value or None if not found.

        Raises:
            HTTPException: If the cookie is missing.
        """
        cookie: str | None = request.cookies.get(self._cookie_name)
        if not cookie:
            raise HTTPException(
                status_code=401,
                detail="Missing Credentials",
            )
        return cookie

    async def __call__(
        self, request: Request, kratos_service: Annotated[KratosService, Depends(depends_kratos_service)]
    ) -> KratosSessionObject:
        """Extract the Kratos session from the request.

        Args:
            request (Request): FastAPI request object.
            kratos_service (KratosService): Kratos service object.

        Returns:
            KratosSessionObject: Kratos session object.

        Raises:
            HTTPException: If the session is invalid.
        """
        cookie: str = self._extract_cookie(request)
        try:
            session: KratosSessionObject = await kratos_service.whoami(cookie_value=cookie)
        except KratosSessionInvalidError as e:
            raise HTTPException(
                status_code=401,
                detail="Invalid Credentials",
            ) from e
        except KratosOperationError as e:
            raise HTTPException(
                status_code=500,
                detail="Internal Server Error",
            ) from e

        return session



"""Package for plugins."""

from enum import StrEnum, auto
from typing import Any


class PluginsEnum(StrEnum):
    """Enumeration for the plugins."""

    OPENTELEMETRY_PLUGIN = auto()
    ODM_PLUGIN = auto()
    HTTPX_PLUGIN = auto()


class PluginState:
    """PluginState represent the state a plugin which to be share to the application and other plugins."""

    def __init__(self, key: str, value: Any) -> None:
        """Initialize the PluginState."""
        self._key: str = key
        self._value = value

    @property
    def key(self) -> str:
        """Get the key."""
        return self._key

    @property
    def value(self) -> Any:
        """Get the value."""
        return self._value


__all__: list[str] = [
    "PluginsEnum",
]



"""Provide PluginManager."""

from .exceptions import (
    InvalidPluginError,
    PluginManagerError,
    PluginPreConditionNotMetError,
)
from .plugin_manager import PluginManager

__all__: list[str] = [
    "PluginManager",
    "PluginManagerError",
    "InvalidPluginError",
    "PluginPreConditionNotMetError",
]
